import numpy as np
import networkx as nx
import math
from epintervene.simobjects import nodestate
import matplotlib.pyplot as plt


class Node:
    def __init__(self, label, generation, state, event_rate):
        self._generation = generation
        self._label = label
        self._state = state
        self._event_rate = event_rate
        self._membership = None

    def infect(self):
        self._state = nodestate.NodeState.INFECTED

    def recover(self):
        self._state = nodestate.NodeState.RECOVERED

    def expose(self):
        self._state = nodestate.NodeState.EXPOSED

    def vaccinate(self):
        self._state = nodestate.NodeState.VACCINATED

    def get_label(self):
        return self._label

    def get_generation(self):
        return self._generation

    def get_state(self):
        return self._state

    def get_event_rate(self):
        return self._event_rate

    def get_membership(self):
        return self._membership

    def set_generation(self, g):
        self._generation = g

    def set_event_rate(self, event_rate):
        self._event_rate = event_rate

    def display_info(self):
        print('Node index: ', self._label, ' state: ', self._state, ' event_rate: ', self._event_rate, ' gen: ',
              self._generation, 'membership: ', self._membership)

    # Give Nodes an option for network class membership, for example in multilayer network or SBM
    # Because Python isn't strongly typed, this can be an int, a float, a string or an Enum
    def set_membership(self, membership):
        self._membership = membership

    def equals(self, node):
        if self._label == node.get_label():
            return True
        else:
            return False


class Edge:
    def __init__(self, left_node, right_node, event_rate):
        self._left_node = left_node  # not just an index, this is a whole Node object
        self._right_node = right_node
        self._event_rate = event_rate

    def infect(self):
        self._right_node.infect()
        self._right_node.set_generation(self._left_node.get_generation() + 1)

    def expose(self):
        self._right_node.expose()
        self._right_node.set_generation(self._left_node.get_generation() + 1)

    def set_event_rate(self, event_rate):
        self._event_rate = event_rate

    def get_event_rate(self):
        return self._event_rate

    def get_right_node(self):
        return self._right_node

    def get_left_node(self):
        return self._left_node

    def display_info(self):
        print('Edge with event rate: ', self._event_rate, ' nodes:')
        self._left_node.display_info()
        self._right_node.display_info()

    def equals(self, other_edge):
        # Imperative to use Node class equality here
        if self._left_node.equals(other_edge._left_node) and self._right_node.equals(other_edge._right_node):
            return True
        else:
            return False


class NetworkBuilder:
    def __init__(self, N):
        self.N = N

    @staticmethod
    def from_degree_distribution(N, degree_dist, return_pos=False):
        number_of_nodes = N * np.array(degree_dist)
        degree_sequence = []
        for i in range(int(math.floor(len(number_of_nodes)))):
            number_with_that_degree = number_of_nodes[i]
            for k in range(int(math.floor(number_with_that_degree))):
                degree_sequence.append(i)
        graphical = nx.is_graphical(degree_sequence)
        if not graphical:
            degree_sequence.append(1)
        G = nx.configuration_model(degree_sequence)
        try:
            G.remove_edges_from(nx.selfloop_edges(G))
        except RuntimeError:
            print('No self loops to remove')
        if return_pos:
            pos = nx.spring_layout(G)
            return G, pos
        return G, None

    @staticmethod
    def from_adjacency_matrix(A, return_pos=False):
        G = nx.from_numpy_matrix(A)
        if return_pos:
            pos = nx.spring_layout(G)
            return G, pos
        return G, None

    @staticmethod
    def erdos_renyi(N, p):
        return nx.erdos_renyi_graph(N, p)

    @staticmethod
    def create_adjacency_list(G):
        #TODO add exception handling for this: ValueError: invalid literal for int() with base 10: '118.0'
        # and demand that node labels must be integers
        adjlist = list(list(map(int, minilist)) for minilist in list(map(str.split, nx.generate_adjlist(G))))

        corrected_adjlist = list(list() for i in range(len(adjlist)))
        for entry in adjlist:
            try:
                corrected_adjlist[entry[0]] = entry
            except IndexError:
                for i in range(entry[0]-len(adjlist) + 1):
                    corrected_adjlist.append(list())
                corrected_adjlist[entry[0]] = entry
                print(entry)

        adjlist = corrected_adjlist
        # make adjlist symmetric:
        len_adj = len(adjlist)
        current_source = 0
        for row in adjlist:
            try:
                source = row[0]
                current_source += 1
            except IndexError:
                row.append(current_source)
                source = row[0]
                current_source += 1
                print(row)
            if len(row) > 1:
                for target in row[1:]:
                    if sum([1 for entry in row if entry==target]) > 1:
                        row.remove(target)
                    try:
                        if source not in adjlist[target]:
                            adjlist[target].append(source)
                    except IndexError:
                        print(source, target)
                        adjlist.append([source])
                # unique_neighbors = list(np.unique(row[1:]))
                # adjlist[-1][1:] = unique_neighbors
        total_entry_count = 0
        for i in range(len_adj):
            if len(adjlist[i]) > 1:
                total_entry_count += len(adjlist[i][1:])
        return adjlist

def visualize(N, graph, pos, gen_collection):
    G = graph
    val_map = {}
    gen_labels_map = {}
    vmax=10
    max_gen = max(gen_collection.keys()) + 1
    for gen in gen_collection.keys():
        nodes = gen_collection[gen]
        for node in nodes:
            val_map[node] = vmax - (gen)
            gen_labels_map.update({node: gen})

    values = [val_map.get(node, 0) for node in G.nodes()]

    nx.draw_networkx_nodes(G, pos=pos, cmap=plt.get_cmap('YlGnBu'), node_color=values, vmin=0, vmax=vmax)
    # nx.draw_networkx_nodes(G, pos=pos, node_color=)
    # gen_labels_map[1]=''
    # gen_labels_map[10]=''
    # gen_labels_map[20]=''
    # gen_labels_map[30]=''
    # nx.draw_networkx_labels(G, pos=pos, with_labels=True, labels=gen_labels_map)
    nx.draw_networkx_edges(G, pos=pos, edge_color='grey', lw=2)

    # V = nx.Graph()
    # V.add_nodes_from([1, 10])
    # V2 = nx.Graph()
    # V2.add_nodes_from([20,30])
    # nx.draw_networkx_nodes(V, pos=pos, node_color='red')
    # nx.draw_networkx_nodes(V2, pos=pos, node_color='orange')
    # nx.draw_networkx_labels(V, pos=pos, with_labels=False, labels={1:'V', 10:'V', 20:'V', 30:'V'})
    plt.axis(False)
    plt.tight_layout()
    plt.show()




import numpy as np
from epintervene.simobjects import network
import matplotlib.pyplot as plt
import networkx as nx
from epintervene.simobjects import simulation
import math


def basic_SIR_simulation():
    print('Basic SIR simulation tutorial')
    # # Create a NetworkBuilder object, to help with creating and configuring the network for the simulation:
    nb = network.NetworkBuilder
    # # Specify a degree distribution to create a configuration model network if desired:
    degree_distrb = binomial_degree_distb(400, 3)
    # # Generating a network from the above degree distribution, with 100 nodes
    G, pos = nb.from_degree_distribution(100, degree_distrb)
    # # If you already have a network, feel free to skip the above steps and generate your own NetworkX object,
    # # or skip this next step and provide a symmetric adjacency list.
    adjlist = nb.create_adjacency_list(G)

    # # Constructing the simulation object
    my_simulation = simulation.Simulation(adj_list=adjlist, N=len(adjlist))

    # # Setting required configurations
    my_simulation.set_uniform_beta(beta=0.9) # .9 people per day per infected person
    my_simulation.set_uniform_gamma(gamma=0.2) # 5 days to recover

    # # Running the simulation
    my_simulation.run_sim()

    # # Obtaining time series results
    time_series_vals, infected_time_series, recovered_time_series = my_simulation.tabulate_continuous_time()
    # # Plotting the results:
    plt.plot(time_series_vals, infected_time_series, label='infected')
    plt.plot(time_series_vals, recovered_time_series, label='recovered')
    plt.legend(loc='upper left')
    plt.title('Auto generated time series values')
    plt.xlabel('Time')
    plt.ylabel('Number of nodes')
    plt.show()

    # # Obtaining time series results with a custom time series end point.
    # # ****TIP: Run at least one simulation without specifying a custom limit, in order to get a sense of what range
    # # of continuous time
    # # the simulation takes to fully run. Then on a second run after experimentation, you can specify a limit.
    time_series_vals, infected_time_series, recovered_time_series = my_simulation.tabulate_continuous_time(
        time_buckets=1000,
        custom_range=True,
        custom_t_lim=15)
    # # Plotting the results:
    plt.plot(time_series_vals, infected_time_series, label='infected')
    plt.plot(time_series_vals, recovered_time_series, label='recovered')
    plt.legend(loc='upper left')
    plt.title('Custom increments of time values')
    plt.xlabel('Time')
    plt.ylabel('Number of nodes')
    plt.show()

    # # Obtaining cumulative infections by generations of infection:
    gen_results = my_simulation.tabulate_generation_results(max_gens=30)

    # # Plotting generation results
    plt.scatter(np.arange(30), gen_results, label='cumulative infections')
    plt.plot(np.arange(30), gen_results)
    plt.title('Cumulative infections by epidemic generations')
    plt.xlabel('Generation number')
    plt.ylabel('Cumulative Infections')
    plt.show()



def basic_SEIR_simulation():
    print('Basic SEIR simulation')
    # # Create a NetworkBuilder object, to help with creating and configuring the network for the simulation:
    nb = network.NetworkBuilder
    # # Specify a degree distribution to create a configuration model network if desired:
    degree_distrb = binomial_degree_distb(400, 3)
    # # Generating a network from the above degree distribution
    G, pos = nb.from_degree_distribution(100, degree_distrb)
    # # If you already have a network, feel free to skip the above steps and generate your own NetworkX object,
    # # or skip this next step and provide a symmetric adjacency list.
    adjlist = nb.create_adjacency_list(G)

    # # Constructing the simulation object
    my_simulation = simulation.SimulationSEIR(N=len(adjlist), adjlist=adjlist)

    # # Setting required configurations
    my_simulation.set_uniform_beta(beta=0.9)
    my_simulation.set_uniform_beta_es(beta=0.45) # Exposed nodes half as infectious as fully infected
    my_simulation.set_uniform_gamma(gamma=0.1) # 10 days to recover
    my_simulation.set_uniform_gamma_ei(gamma=0.2) # 5 days to develop symptoms

    # # Running the simulation
    my_simulation.run_sim(wait_for_recovery=True) # Setting wait for recovery to True, to observe the full epidemic curve

    # # Obtaining time series results
    time_series_vals, infected_time_series, recovered_time_series, exposed_time_series = my_simulation.tabulate_continuous_time()

    # # Plotting the results:
    plt.plot(time_series_vals, infected_time_series, label='infected')
    plt.plot(time_series_vals, recovered_time_series, label='recovered')
    plt.plot(time_series_vals, exposed_time_series, label='exposed')
    plt.legend(loc='upper left')
    plt.title('Auto generated time series values')
    plt.xlabel('Time')
    plt.ylabel('Number of nodes')
    plt.show()

    # # Obtaining time series results with a custom time series end point.
    # # ****TIP: Run at least one simulation without specifying a custom limit, in order to get a sense of what range
    # # of continuous time
    # # the simulation takes to fully run. Then on a second run after experimentation, you can specify a limit.
    time_series_vals, infected_time_series, recovered_time_series, exposed_time_series = my_simulation.tabulate_continuous_time(
        time_buckets=1000,
        custom_range=True,
        custom_t_lim=15)
    # # Plotting the results:
    plt.plot(time_series_vals, infected_time_series, label='infected')
    plt.plot(time_series_vals, recovered_time_series, label='recovered')
    plt.plot(time_series_vals, exposed_time_series, label='exposed')
    plt.legend(loc='upper left')
    plt.title('Custom increments of time values')
    plt.xlabel('Time')
    plt.ylabel('Number of nodes')
    plt.show()

    # # Obtaining cumulative infections by generations of infection:
    gen_results = my_simulation.tabulate_generation_results(max_gens=30)

    # # Plotting generation results
    plt.scatter(np.arange(30), gen_results, label='cumulative infections')
    plt.plot(np.arange(30), gen_results)
    plt.title('Cumulative infections by epidemic generations')
    plt.xlabel('Generation number')
    plt.ylabel('Cumulative Infections')
    plt.show()


def basic_SIR_groups_simulation():
    print('Basic SIR simulation with membership groups')
    # Creating a network from a Stochastic Block Model
    nb = network.NetworkBuilder
    G, pos, _ = create_zoo_stochastic_block_model(tiger_elephant_block=.05, tiger_bird_block=.03,
                                                  bird_elephant_block=.02)
    adjlist = nb.create_adjacency_list(G)

    # # Assigning indices for the split population
    tiger_population = int(len(adjlist) / 3)
    bird_population = int(len(adjlist) / 3)
    elephant_population = int(len(adjlist) / 3)

    # # Creating a vector of node memberships, indexed by node label (index from above)
    node_membership_vector = []
    for i in range(tiger_population):
        node_membership_vector.append('tiger')
    for j in range(bird_population):
        node_membership_vector.append('bird')
    for k in range(elephant_population):
        node_membership_vector.append('elephant')

    # # Setting up the simulation with memberships
    sim = simulation.Simulation(N=len(adjlist), adj_list=adjlist, membership_groups=['tiger', 'bird', 'elephant'],
                                node_memberships=node_membership_vector)

    # # Configuring the disease parameters
    sim.set_uniform_beta(0.7)
    sim.set_uniform_gamma(0.21)

    # # Running the simulation, must specify to track memberships
    sim.run_sim(with_memberships=True, wait_for_recovery=True, uniform_rate=True)

    # # Recording results with group memberships
    ts, membership_ts_infc, membership_ts_rec = sim.tabulate_continuous_time_with_groups(time_buckets=1000, custom_range=True,
                                                                      custom_t_lim=30)

    # # Plotting and interpreting the results
    plt.figure(0)
    for group in membership_ts_infc.keys():
        plt.plot(ts, membership_ts_infc[group], label=f'{group} infected')
        plt.plot(ts, membership_ts_rec[group], label=f'{group} recovered', ls='--')
    plt.xlabel('Time t')
    plt.ylabel('Number of nodes infected in network group')
    plt.legend(loc='upper left')
    plt.title('SIR Continuous time results with group membership')
    plt.box(on=False)
    plt.show()

    ts, infect_ts, recover_ts = sim.tabulate_continuous_time(time_buckets=1000, custom_range=True, custom_t_lim=30)
    plt.figure(1)
    plt.plot(ts, infect_ts, color='blue', label='Infected')
    plt.plot(ts, recover_ts, color='green', label='Recovered')
    plt.xlabel('Time t')
    plt.ylabel('Number of nodes in class')
    plt.title('SIR Continuous time results \n(total population, without showing group membership)')
    plt.legend(loc='upper left')
    plt.box(on=False)
    # plt.show()

    ts_by_gen = sim.tabulate_generation_results(20)
    plt.figure(2)
    plt.plot(np.arange(len(ts_by_gen)), ts_by_gen)
    plt.scatter(np.arange(len(ts_by_gen)), ts_by_gen)
    plt.xlabel('Generation number')
    plt.ylabel('Cumulative infections by generation')
    plt.title('SIR Cumulative Epidemic Generation size results')
    plt.box(on=False)
    plt.show()


def basic_SEIR_groups_simulation():
    print('Basic SEIR simulation with membership groups')
    # Creating a network from a Stochastic Block Model
    nb = network.NetworkBuilder
    G, pos, _ = create_zoo_stochastic_block_model(tiger_elephant_block=.05, tiger_bird_block=.03,
                                                  bird_elephant_block=.02)
    adjlist = nb.create_adjacency_list(G)

    # # Assigning indices for the split population
    tiger_population = int(len(adjlist) / 3)
    bird_population = int(len(adjlist) / 3)
    elephant_population = int(len(adjlist) / 3)

    # # Creating a vector of node memberships, indexed by node label (index from above)
    node_membership_vector = []
    for i in range(tiger_population):
        node_membership_vector.append('tiger')
    for j in range(bird_population):
        node_membership_vector.append('bird')
    for k in range(elephant_population):
        node_membership_vector.append('elephant')

    # # Setting up the simulation with memberships
    sim = simulation.SimulationSEIR(N=len(adjlist), adjlist=adjlist, node_memberships=node_membership_vector,
                                    membership_groups=['tiger', 'bird', 'elephant'])

    # # Configuring the disease parameters
    sim.set_uniform_gamma(0.1)
    sim.set_uniform_beta(0.5)
    sim.set_uniform_beta_es(0.25)
    sim.set_uniform_gamma_ei(0.2)

    # # Running the simulation
    sim.run_sim(with_memberships=True, wait_for_recovery=True, uniform_rate=True)

    # # Recording the results
    ts, membership_ts_infc, membership_ts_exp, membership_ts_rec = sim.tabulate_continuous_time_with_groups()

    # # Plotting and interpreting the results
    plt.figure(0)
    colors = {'tiger': 'blue', 'bird': 'red', 'elephant': 'orange'}
    for group in membership_ts_infc.keys():
        plt.plot(ts, membership_ts_infc[group], label=f'{group} infected', color=colors[group])
        plt.plot(ts, membership_ts_exp[group], label=f'{group} exposed', ls='--', color=colors[group])
        plt.plot(ts, membership_ts_rec[group], label=f'{group} recovered', ls=':', color=colors[group])
    plt.xlabel('Time t')
    plt.ylabel('Number of nodes infected in network group')
    plt.legend(loc='upper right')
    plt.title('SEIR Continuous time results with group membership')
    plt.show()

    ts, infect_ts, recover_ts, exposed_ts = sim.tabulate_continuous_time(1000, custom_range=True, custom_t_lim=30)

    plt.figure(1)
    plt.plot(ts, infect_ts, color='red', label='Infected')
    plt.plot(ts, exposed_ts, color='orange', label='Exposed')
    plt.plot(ts, recover_ts, color='green', label='Recovered')
    plt.xlabel('Time t')
    plt.ylabel('Number of nodes in class')
    plt.legend(loc='upper left')
    plt.title('SEIR Continuous time results (total population, without showing group membership)')
    # plt.show()

    ts_by_gen = sim.tabulate_generation_results(max_gens=20)
    plt.figure(2)
    plt.plot(np.arange(len(ts_by_gen)), ts_by_gen)
    plt.scatter(np.arange(len(ts_by_gen)), ts_by_gen)
    plt.xlabel('Generation number')
    plt.ylabel('Cumulative infections by generation')
    plt.title('SEIR Cumulative Epidemic Generation size results')
    plt.show()


def create_zoo_stochastic_block_model(N=300, tiger_bird_block=0.01, bird_elephant_block=0.02,
                                      tiger_elephant_block=0.005):
    A = np.zeros((N, N))
    tiger_block = 10 / N
    bird_block = 10 / N
    elephant_block = 20 / N
    tiger_bird_block = tiger_bird_block / N
    bird_elephant_block = bird_elephant_block / N
    tiger_elephant_block = tiger_elephant_block / N

    tiger_pop = int(len(A) / 3)
    bird_pop = int(len(A) / 3)
    elephant_pop = int(len(A) / 3)

    tiger_range = np.arange(0, tiger_pop)
    bird_range = np.arange(tiger_pop, tiger_pop + bird_pop)
    elephant_range = np.arange(tiger_pop + bird_pop, tiger_pop + bird_pop + elephant_pop)

    for i in tiger_range:
        for j in range(i, tiger_range[-1]):
            A[i, j] = 1 if np.random.rand() < tiger_block else 0
            A[j, i] = A[i, j]
        for j in bird_range:
            A[i, j] = 1 if np.random.rand() < tiger_bird_block else 0
            A[j, i] = A[i, j]
        for j in elephant_range:
            A[i, j] = 1 if np.random.rand() < tiger_elephant_block else 0
            A[j, i] = A[i, j]
    for i in bird_range:
        for j in range(i, bird_range[-1]):
            A[i, j] = 1 if np.random.rand() < bird_block else 0
            A[j, i] = A[i, j]
        for j in elephant_range:
            A[i, j] = 1 if np.random.rand() < bird_elephant_block else 0
            A[j, i] = A[i, j]
    for i in elephant_range:
        for j in range(i, elephant_range[-1]):
            A[i, j] = 1 if np.random.rand() < elephant_block else 0
            A[j, i] = A[i, j]

    np.fill_diagonal(A, 0)
    G, pos = network.NetworkBuilder.from_adjacency_matrix(A, return_pos=True)

    node_colors = {}
    for n in G.nodes():
        if n in tiger_range:
            node_colors[n] = 'blue'
        elif n in bird_range:
            node_colors[n] = 'red'
        elif n in elephant_range:
            node_colors[n] = 'orange'
    nx.draw_networkx_nodes(G, pos, G.nodes(), node_color=node_colors.values(), node_size=20)
    nx.draw_networkx_edges(G, pos)
    plt.title('Tigers (blue), Bird (red), Elephant (orange) network')
    plt.box(on=False)
    # plt.show()
    return G, pos, A


def power_law_degree_distrb(maxk=40, alpha=2, mu=5):
    p_k = np.empty(maxk)
    p_k[0] = 0
    for k in range(1, maxk):
        p_k[k] = (k ** (-alpha)) * (math.e ** (-k / mu))
    p_k = p_k / np.sum(p_k)
    return p_k


def binomial_degree_distb(N, lam=6):
    p_k = np.empty(N)
    p = lam / N
    for k in range(0, len(p_k)):
        p_k[k] = (p ** k) * ((1 - p) ** (N - k)) * math.comb(N, k)
    return p_k


if __name__ == '__main__':
    basic_SIR_simulation()
    basic_SEIR_simulation()
    basic_SIR_groups_simulation()
    basic_SEIR_groups_simulation()



import random

import numpy as np

from epintervene.simobjects import eventtype
from epintervene.simobjects import network
from epintervene.simobjects import nodestate


class Simulation:
    """
    An SIR simulation object storing network information, rate parameters, and simulation run results.

    """

    def __init__(self, N, adj_list=None, max_unitless_sim_time=1000000, membership_groups=None, node_memberships=None):
        """
        Create a single instance of an SIR Simulation object.

        :param adj_matrix: Input array consisting of the adjacency matrix for the underlying network.
        :param max_unitless_sim_time: Max value the event-driven simulation will cease running at when time draw reaches this value.
        :param membership_groups: Tags for membership groups or meta-populations in underlying network model. Cqn be any type.
        :param node_memberships: Ordered list of membership group values where list index corresponds to node label in the adjacency matrix/list.
        """
        self.total_sim_time = max_unitless_sim_time
        self._current_sim_time = 0
        self._adjlist = adj_list
        self._N = N
        self._uniform_beta = None
        self._uniform_gamma = None
        self._out_degree_IS_events = {}
        self._out_degree_IS_lengths = {}
        self._in_degree_IS_events = {}
        self._recovered_nodes = []
        self._recovery_events = {}
        self._recovery_events_keys = []
        self._highest_gen = 0
        self._gen_collection = {}
        self._gen_collection_active = {}
        self._active_node_dict = {}
        self._total_num_timesteps = 0
        self._time_series = [0]
        self._real_time_srs_infc = []
        self._real_time_srs_rec = []
        self.active_gen_ts = []
        self.total_gen_ts = []
        self._generational_emergence = {0: 0}
        self._membership_groups = membership_groups
        self._membership_time_series_infc = {0: []}
        self._membership_time_series_rec = {0: []}
        self.track_memberships = False
        self.use_uniform_rate = True
        self._node_memberships = node_memberships
        self.infectious_degree_counter = []
        self._current_number_IS_events = 0
        self._current_number_recovery_events = 0
        self._max_num_recovery_events = 0
        self._potential_IS_events_flat = []
        self._edge_keys = {}
        self._next_key = 0
        # Idea here is a matrix, where it is a list of lists. Each row will be gens 1,100, the sizes at that time series t
        # When it's time to tally, should use the built-in tallying function to bin into even matrix
        self._gens_size_over_time = None
        # In order to compute things to most efficiently, need to have the leading row of active gens counts and then just append or remove
        #  from that count in order not to have to compute the size at every generation
        self._current_active_gen_sizes = np.zeros(100)
        self._gen_sizes_function_of_gen = []

    def get_gen_collection(self):
        return self._gen_collection

    def get_gen_collection_active(self):
        return self._gen_collection_active

    def get_generational_emergence(self):
        return self._generational_emergence

    def get_time_series_vals(self):
        return self._time_series

    def get_real_time_series_infected(self):
        return self._real_time_srs_infc

    def get_real_time_series_recovered(self):
        return self._real_time_srs_rec

    def get_membership_time_series_infected(self):
        return self._membership_time_series_infc

    def get_membership_time_series_recovered(self):
        return self._membership_time_series_rec

    def set_adjlist(self, adjlist):
        self._adjlist = adjlist

    def set_uniform_beta(self, beta):
        self._uniform_beta = beta

    def set_uniform_gamma(self, gamma):
        self._uniform_gamma = gamma

    def _reset_node_states(self):
        # WARNING: Not to be used outside of designated purpose. Not sufficient for re-starting a simulation.
        self._out_degree_IS_events = {}
        self._out_degree_IS_lengths = {}
        self._in_degree_IS_events = {}
        self._current_number_IS_events = 0
        self._potential_IS_events_flat = []
        self._edge_keys = {}
        self._next_key = 0

    def _initialize_patient_zero(self, label=None):
        if label is None:
            try:
                p_zero_idx = random.randint(0, self._N - 1)
            except ValueError:
                print(self._N)
        elif label is not None:
            p_zero_idx = label
        if self._uniform_gamma is None:
            raise AttributeError(self._uniform_gamma,
                                 'Please provide a recovery rate gamma to the simulation via the constructor or the '
                                 'set_uniform_gamma method')
        if self._uniform_beta is None:
            raise AttributeError(self._uniform_beta,
                                 'Please provide an infection rate beta to the simulation via the constructor or the '
                                 'set_uniform_beta method')
        if self.use_uniform_rate:
            patient_zero = network.Node(p_zero_idx, 0, nodestate.NodeState.INFECTED, event_rate=self._uniform_gamma)
        else:
            patient_zero = network.Node(p_zero_idx, 0, nodestate.NodeState.INFECTED, event_rate=self._Gamma[p_zero_idx])
        if self.track_memberships:
            patient_zero.set_membership(self._node_memberships[p_zero_idx])
        self._active_node_dict[p_zero_idx] = patient_zero
        self._gen_collection[0] = [p_zero_idx]
        self._gen_collection_active[0] = [p_zero_idx]
        self.active_gen_ts.append(1)
        self.total_gen_ts.append(1)
        self._current_active_gen_sizes[0] = 1
        self._gen_sizes_function_of_gen.append(list(self._current_active_gen_sizes))
        self._recovery_events[p_zero_idx] = patient_zero
        self._recovery_events_keys.append(p_zero_idx)
        self._current_number_recovery_events = 1
        self._out_degree_IS_events[p_zero_idx] = []
        self._out_degree_IS_lengths[p_zero_idx] = 0
        self._add_IS_events(patient_zero)

    def run_sim(self, with_memberships=False, uniform_rate=True, wait_for_recovery=False, visualize=False,
                viz_graph=None, viz_pos=None, p_zero=None, kill_by=None, record_active_gen_sizes=False):
        """
        Main method for running a single realization of the epidemic simulation.

        :param with_memberships: Set to True if specified memberships in Simulation configuration
        :param uniform_rate: Will use a uniform transmission and recovery event probabilites instead of a customized matrix. Increases simulation speed, use if possible.
        :param wait_for_recovery: Will run simulation until all possible nodes have recovered.
        """
        self.use_uniform_rate = uniform_rate
        if self.use_uniform_rate and (
                self._uniform_beta is None or self._uniform_gamma is None):
            raise Exception(
                'Since uniform_rate=True, must set all transmission and recovery parameter values via set_uniform_{param} method}')
        if with_memberships:
            self.track_memberships = True
        if self.track_memberships:
            self._init_membership_state_time_series()
        self._initialize_patient_zero(label=p_zero)
        self._gens_size_over_time = []
        self._gens_size_over_time.append(list(self._current_active_gen_sizes))
        while self._current_sim_time < self.total_sim_time:
            # Run one step
            self._single_step(uniform_rate=uniform_rate, visualize=visualize, viz_graph=viz_graph, viz_pos=viz_pos,
                              record_active_gen_sizes=record_active_gen_sizes)

            self._total_num_timesteps += 1
            if self._current_number_IS_events == 0:
                if not wait_for_recovery:
                    break
                elif len(self._recovery_events) == 0:
                    break
            if kill_by is not None and self._highest_gen >= kill_by:
                should_quit = self._check_active_gens(kill_by)
                if should_quit:
                    break

    def _check_active_gens(self, kill_by):
        for i in range(0, kill_by):
            if len(self._gen_collection_active[i]) > 0:
                return False
        return True

    def _single_step(self, visualize=False, uniform_rate=True, viz_graph=None, viz_pos=None,
                     record_active_gen_sizes=False):
        # Note: record_active_gen_sizes makes the simulation twice as slow, so use with caution and only use for results needing ensembleing
        # TODO could probably do the same with active gen sizes! Make it quicker for the generational results later
        self.use_uniform_rate = uniform_rate
        if visualize:
            self._visualize_network(viz_graph, viz_pos)
        event_class, next_event, tau = self.draw_event_class()
        self._current_sim_time += tau
        self._time_series.append(self._time_series[-1] + tau)
        self._real_time_srs_infc.append(
            self._current_number_recovery_events)  # should this actually happen before the time is incremented? Like the time keeping
        self._real_time_srs_rec.append(len(self._recovered_nodes))
        if record_active_gen_sizes:
            self._gens_size_over_time.append(list(self._current_active_gen_sizes))
        if self.track_memberships:
            self._record_membership_states()

        if event_class is not None: #TODO is this infecting vax'ed people? start here tuesday
            if event_class == eventtype.EventType.INFECTEDSUSCEPTIBLE:
                infection_event = next_event
                if next_event.get_right_node().get_state()==nodestate.NodeState.SUSCEPTIBLE:
                    infection_event.infect()
                    self._recovery_events[infection_event.get_right_node().get_label()] = infection_event.get_right_node()
                    self._recovery_events_keys.append(infection_event.get_right_node().get_label())
                    self._current_number_recovery_events += 1
                    self._max_num_recovery_events += 1

                    try:  # If they are a member of an existing generation, bookkeeping will happen here:
                        self._gen_collection[infection_event.get_right_node().get_generation()].append(
                            infection_event.get_right_node().get_label())
                        self._gen_collection_active[infection_event.get_right_node().get_generation()].append(
                            infection_event.get_right_node().get_label())
                        self.active_gen_ts.append(self.active_gen_ts[-1])  # Active gens stays the same this round
                        self.total_gen_ts.append(self.total_gen_ts[-1])  # Total gens stays the same this round
                        try:
                            self._current_active_gen_sizes[
                                infection_event.get_right_node().get_generation()] += 1  # One more active member of the generation
                        except IndexError:
                            pass # Don't record gens larger than 100
                    except KeyError:  # If they are the first member of a new generation, book keeping happens here
                        self._gen_collection[infection_event.get_right_node().get_generation()] = [
                            infection_event.get_right_node().get_label()]
                        self._gen_collection_active[infection_event.get_right_node().get_generation()] = [
                            infection_event.get_right_node().get_label()]
                        self._highest_gen += 1
                        self._generational_emergence[self._highest_gen] = self._current_sim_time
                        self.active_gen_ts.append(self.active_gen_ts[-1] + 1)  # Active gens increases by one
                        self.total_gen_ts.append(self.total_gen_ts[-1] + 1)  # Total gens increases by one
                        try:
                            self._current_active_gen_sizes[
                                infection_event.get_right_node().get_generation()] += 1  # One more active member of the generation
                        except IndexError:
                            pass # Don't record generations larger than 100
                        self._gen_sizes_function_of_gen.append(list(self._current_active_gen_sizes))
                    self._update_IS_events(infection_IS_event=infection_event)
                    self._add_IS_events(infection_event.get_right_node())
            if event_class == eventtype.EventType.RECOVER:
                recovery_event = next_event
                recovery_event.recover()
                self._recovery_events.pop(recovery_event.get_label())
                self._current_number_recovery_events -= 1
                self._update_IS_events(recovery_event=recovery_event)
                self._recovered_nodes.append(recovery_event)
                try:
                    self._current_active_gen_sizes[
                        recovery_event.get_generation()] -= 1  # One less active member of the generation
                except IndexError:
                    pass # Don't record gens greater than 100
                try:
                    self._gen_collection_active[recovery_event.get_generation()].remove(recovery_event.get_label())
                except ValueError:
                    pass
                if len(self._gen_collection_active[
                           recovery_event.get_generation()]) == 0:  # If this was the last remaining member of the generation
                    if self.active_gen_ts[-1] == 0:
                        self.active_gen_ts.append(
                            self.active_gen_ts[-1])  # If active generations is already 0, then stays at 0
                    else:
                        self.active_gen_ts.append(
                            self.active_gen_ts[-1] - 1)  # Otherwise, reduces active generations by 1
                else:
                    self.active_gen_ts.append(self.active_gen_ts[-1])
                self.total_gen_ts.append(self.total_gen_ts[-1])  # this always should stay the same no matter what

    def _update_IS_events(self, infection_IS_event=None, recovery_event=None):
        if infection_IS_event is not None:
            try:
                in_degree_events = self._in_degree_IS_events[infection_IS_event.get_right_node().get_label()]
                for event in in_degree_events:
                    left_node_idx = event.get_left_node().get_label()
                    try:
                        self._out_degree_IS_events[left_node_idx].remove(event)
                        self._out_degree_IS_lengths[left_node_idx] -= 1
                        self._current_number_IS_events -= 1
                        flat_key = self._edge_keys[event.get_left_node().get_label()][
                            event.get_right_node().get_label()]
                        self._potential_IS_events_flat[flat_key] = None
                        if self._out_degree_IS_lengths[left_node_idx] == 0:
                            self._gen_collection_active[infection_IS_event.get_left_node().get_generation()].remove(
                                infection_IS_event.get_left_node().get_label())
                    except ValueError:
                        pass
                self._in_degree_IS_events[infection_IS_event.get_right_node().get_label()] = []
            except KeyError:
                pass
        elif recovery_event is not None:
            try:
                out_degree_events = self._out_degree_IS_events[recovery_event.get_label()]
                for event in out_degree_events:
                    flat_key = self._edge_keys[event.get_left_node().get_label()][event.get_right_node().get_label()]
                    try:
                        self._potential_IS_events_flat[flat_key] = None
                    except:
                        print(flat_key)
                self._out_degree_IS_events[recovery_event.get_label()] = []
                self._out_degree_IS_lengths[recovery_event.get_label()] = 0
                self._current_number_IS_events -= len(out_degree_events)
            except KeyError:
                pass
            try:
                self._in_degree_IS_events[recovery_event.get_label()] = []
            except KeyError:
                pass

    def _add_IS_events(self, infected_node):
        infection_adjlist = self._adjlist[infected_node.get_label()]
        adjlist_length = len(infection_adjlist)
        if adjlist_length > 1:
            infected_label = infected_node.get_label()
            length = adjlist_length
            for n in range(1, length):
                j = infection_adjlist[n]
                if self.use_uniform_rate:
                    candidate_node = network.Node(j, -1, nodestate.NodeState.SUSCEPTIBLE, self._uniform_gamma)
                else:
                    print('Must use Uniform Rate (set Simulation.use_uniform_rate to True)')
                    candidate_node = network.Node(j, -1, nodestate.NodeState.SUSCEPTIBLE, self._Gamma[j])
                if self.track_memberships:
                    candidate_node.set_membership(self._node_memberships[candidate_node.get_label()])
                neighbor_node = self._existing_node(candidate_node)
                if neighbor_node.get_state() == nodestate.NodeState.SUSCEPTIBLE:
                    if self.use_uniform_rate:
                        edge_ij = network.Edge(infected_node, neighbor_node, self._uniform_beta)
                    else:
                        print('Must use Uniform Rate (set Simulation.use_uniform_rate to True)')
                        edge_ij = network.Edge(infected_node, neighbor_node, self._Beta[infected_label][j])
                    try:
                        self._out_degree_IS_events[infected_node.get_label()].append(edge_ij)
                        self._out_degree_IS_lengths[infected_node.get_label()] += 1
                    except KeyError:
                        self._out_degree_IS_events[infected_node.get_label()] = [edge_ij]
                        self._out_degree_IS_lengths[infected_node.get_label()] = 1
                    try:
                        self._in_degree_IS_events[neighbor_node.get_label()].append(edge_ij)
                    except KeyError:
                        self._in_degree_IS_events[neighbor_node.get_label()] = [edge_ij]
                    self._current_number_IS_events += 1
                    # Dict of Dicts which holds the "key" to the flat list
                    self._potential_IS_events_flat.append(edge_ij)
                    # the length of the above list should hopefully be the same as current number of events
                    try:
                        self._edge_keys[edge_ij.get_left_node().get_label()][
                            edge_ij.get_right_node().get_label()] = self._next_key
                    except KeyError:
                        self._edge_keys[edge_ij.get_left_node().get_label()] = {
                            edge_ij.get_right_node().get_label(): self._next_key}
                    self._next_key += 1

    def _existing_node(self, candidate_node):
        if candidate_node.get_label() in self._active_node_dict.keys():
            return self._active_node_dict[candidate_node.get_label()]
        else:
            self._active_node_dict[candidate_node.get_label()] = candidate_node
        return candidate_node

    def _edge_list_contains(self, edge):
        # for e in self._potential_IS_events._event_list:
        left_node_label = edge.get_left_node().get_label()
        for e in self._out_degree_IS_events[left_node_label]:
            # for e in self._potential_IS_events._event_list:
            if e.equals(edge):
                return True
        right_node_label = edge.get_right_node().get_label()
        for e in self._in_degree_IS_events[right_node_label]:
            # TODO don't know if we really need this but its a just in case
            if e.equals(edge):
                return True
        return False

    def _record_membership_states(self):
        for group in self._membership_groups:
            infected_in_group = list(node for node in self._active_node_dict.values() if node.get_state() == nodestate.NodeState.INFECTED and node.get_membership() == group)
            self._membership_time_series_infc[group].append(len(infected_in_group))
            recovered_in_group = list(node for node in self._active_node_dict.values() if
                                     node.get_state() == nodestate.NodeState.RECOVERED and node.get_membership() == group)
            self._membership_time_series_rec[group].append(len(recovered_in_group))

    def _init_membership_state_time_series(self):
        self._membership_time_series_infc = {}
        for group in self._membership_groups:
            self._membership_time_series_infc[group] = []
            self._membership_time_series_rec[group] = []

    def _visualize_network(self, viz_graph, viz_pos):
        network.visualize(self._N, viz_graph, viz_pos, self._gen_collection)
        print('In progress')

    def draw_event_class(self):
        partition_end_markers = {}
        num_of_recovery_events = self._current_number_recovery_events
        num_of_infection_possible_events = self._current_number_IS_events
        partition_end_markers[0] = num_of_infection_possible_events * self._uniform_beta
        partition_end_markers[1] = num_of_infection_possible_events * self._uniform_beta + num_of_recovery_events * self._uniform_gamma
        random_draw = random.uniform(0, partition_end_markers[1])
        sum_of_rates = (self._uniform_beta * num_of_infection_possible_events + self._uniform_gamma * num_of_recovery_events)
        tau = random.expovariate(max(sum_of_rates, .0000001))
        if random_draw < partition_end_markers[0]:
            found_next_event = None
            while found_next_event is None:
                random_idx = int(random.uniform(0, self._next_key))
                found_next_event = self._potential_IS_events_flat[random_idx]
            return eventtype.EventType.INFECTEDSUSCEPTIBLE, found_next_event, tau
        else:
            found_next_event = None
            while found_next_event is None:
                try:
                    random_idx = random.randint(0, self._max_num_recovery_events)
                except ValueError:
                    print('Value error')
                    print(f'number recovery nodes {print(len(self._recovery_events))}')
                next_event_label = self._recovery_events_keys[random_idx]
                try:
                    found_next_event = self._recovery_events[next_event_label]
                except KeyError:
                    continue
            return eventtype.EventType.RECOVER, found_next_event, tau

    def tabulate_generation_results(self, max_gens):
        """
        Compile the results from the simulation in terms of epidemic generations.

        :param max_gens: Maximum generations desired to tabulate and return information for
        :return: Array indexed by generations from 0 to max_gens of how many cumulative infected nodes there are in generations less than or equal to the index
        """
        gens = max_gens
        infection_time_srs_by_gen = np.zeros(gens)
        s = 1
        m = 1
        s_max = 1
        for gen in range(max_gens):  # {0: 1, 1: 12, 14, 2: 16, 42, ....
            infection_time_srs_by_gen[gen] = s
            try:
                m = len(self._gen_collection[gen + 1])  # num infected in gen g
                s += m
                s_max = s
            except KeyError:
                # make m=0 and s=the last s for the rest of the "time series"
                s = s_max
                m = 0
        return infection_time_srs_by_gen

    def tabulate_continuous_time(self, time_buckets=100, custom_range=False, custom_t_lim=100, active_gen_info=False,
                                 active_gen_sizes=False, active_gen_by_gen=False):
        """
        Compile the results from the simulation as a time series.

        :param time_buckets: How many ticks to partition the total time into
        :param custom_range: Defaults to False, set to True if desire a set maximum time value (suggested if collecting an ensemble of results)
        :param custom_t_lim: If custom_range is True, provide max time value for the time series
        :return: three arrays: time series values, number of infections time series, number recovered time series
        """
        max_time = max(self._time_series)
        time_partition = np.linspace(0, max_time + 1, time_buckets, endpoint=False)
        if custom_range:
            time_partition = np.linspace(0, custom_t_lim, time_buckets, endpoint=False)
        infection_time_series = np.zeros(len(time_partition))
        recover_time_series = np.zeros(len(time_partition))
        active_gen_time_series = np.zeros(len(time_partition))
        total_gen_time_series = np.zeros(len(time_partition))
        active_gen_sizes_ts = np.zeros((len(time_partition), 100))
        if active_gen_sizes:
            self._gens_size_over_time = np.array(self._gens_size_over_time)
        ts_length = len(self._time_series)

        idx_floor = 0
        for t in range(ts_length - 1):
            real_time_val = self._time_series[t]
            real_time_infected = self._real_time_srs_infc[t]
            real_time_recovered = self._real_time_srs_rec[t]
            if active_gen_info:
                real_time_active_gens = self.active_gen_ts[t]
                real_time_total_gens = self.total_gen_ts[t]
            if active_gen_sizes:
                real_time_gen_sizes = self._gens_size_over_time[t]
            found_soonest = False
            idx = idx_floor
            while not found_soonest and idx<time_buckets:
                upper_val = time_partition[idx]
                if real_time_val <= upper_val:
                    idx_floor = idx
                    infection_time_series[idx:] = real_time_infected
                    recover_time_series[idx:] = real_time_recovered
                    if active_gen_info:
                        active_gen_time_series[idx:] = real_time_active_gens
                        total_gen_time_series[idx:] = real_time_total_gens
                    if active_gen_sizes:
                        active_gen_sizes_ts[idx:] = real_time_gen_sizes
                    found_soonest = True
                idx += 1
        if active_gen_info and active_gen_sizes and active_gen_by_gen:
            return time_partition, infection_time_series, recover_time_series, \
                   active_gen_time_series, total_gen_time_series, active_gen_sizes_ts, np.array(self._gen_sizes_function_of_gen)
        if active_gen_info and active_gen_sizes:
            return time_partition, infection_time_series, recover_time_series, \
                   active_gen_time_series, total_gen_time_series, active_gen_sizes_ts
        elif active_gen_info:
            return time_partition, infection_time_series, recover_time_series, \
                   active_gen_time_series, total_gen_time_series
        return time_partition, infection_time_series, recover_time_series

    def tabulate_continuous_time_with_groups(self, time_buckets=100, custom_range=False, custom_t_lim=100):
        """
        Compile the results of the simulation as a dictionary of time series, yielding a distinct time series for
        each membership group.

        :param time_buckets: How many ticks to partition the total time into
        :param custom_range: Defaults to False, set to True if desire a set maximum time value (suggested if collecting an ensemble of results)
        :param custom_t_lim: If custom_range is True, provide max time value for the time series
        :return: two objects: the time series values, and a dict of infection time series keyed by membership group
        """
        max_time = max(self._time_series)
        time_partition = np.linspace(0, max_time + 1, time_buckets, endpoint=False)
        if custom_range:
            time_partition = np.linspace(0, custom_t_lim, time_buckets, endpoint=False)
        infection_time_series_group_dict = {}
        recover_time_series_group_dict = {}
        for group in self._membership_groups:
            infection_time_series_group_dict[group] = np.zeros(len(time_partition))
            recover_time_series_group_dict[group] = np.zeros(len(time_partition))
        ts_length = len(self._time_series)

        idx_floor = 0
        for t in range(ts_length - 1):
            real_time_val = self._time_series[t]
            found_soonest = False
            idx = idx_floor
            while not found_soonest and idx<time_buckets:
                upper_val = time_partition[idx]
                if real_time_val <= upper_val:
                    idx_floor = idx
                    for group in self._membership_groups:
                        real_time_infected = self._membership_time_series_infc[group][t]
                        real_time_recovered = self._membership_time_series_rec[group][t]
                        infection_time_series_group_dict[group][idx:] = real_time_infected
                        recover_time_series_group_dict[group][idx:] = real_time_recovered
                    found_soonest = True
                idx += 1

        return time_partition, infection_time_series_group_dict, recover_time_series_group_dict


class SimulationSEIR(Simulation):
    """
    An SEIR simulation object storing network information, rate parameters, and simulation run results.

    """

    def __init__(self, N, adjlist=None, max_unitless_sim_time=1000000, membership_groups=None, node_memberships=None):
        """
        Create a single instance of an SEIR Simulation object.

        :param N: Number of nodes in the network
        :param adjmatrix: Adjacency matrix, numpy array, can be None
        :param adjlist: Adjacency list, must be symmetric. Nested list should have entries of form e.g. [i j k] where node i has neighbors j and k. New source node is always the left-most node of a new line.
        :param max_unitless_sim_time: Max value the event-driven simulation will cease running at when time draw reaches this value.
        :param membership_groups:  Tags for membership groups or meta-populations in underlying network model. Cqn be any type.
        :param node_memberships: Ordered list of membership group values where list index corresponds to node label in the adjacency matrix/list.
        """
        super().__init__(adj_list=adjlist, N=N, max_unitless_sim_time=max_unitless_sim_time,
                         membership_groups=membership_groups, node_memberships=node_memberships)
        self._current_exposed = []
        self._current_number_ES_events = 0
        self._current_number_EI_events = 0
        self._potential_ES_events_flat = []
        self._edge_keys_es = {}
        self._next_key_es = 0
        self._real_time_srs_exp = []
        self._cont_time_exposed_dict = {}
        self._membership_time_series_exp = {0: []}
        self._in_degree_ES_events = {}
        self._out_degree_ES_events = {}
        self._out_degree_ES_lengths = {}
        self._uniform_gamma_ei = None
        self._uniform_beta_es = None
        self._exposed_to_infected_events = {}
        self._exposed_to_infected_events_keys = []
        self._max_number_EI_events = 0
        # TODO deal with generations of exposed emergence as well
        # should probably also determine in the model if there's completely asymptomatic cases as well (E->R events)

    def set_uniform_beta_es(self, beta):
        self._uniform_beta_es = beta

    def set_uniform_gamma_ei(self, gamma):
        self._uniform_gamma_ei = gamma

    def run_sim(self, with_memberships=False, uniform_rate=True, wait_for_recovery=False, visualize=False,
                viz_graph=None, viz_pos=None, p_zero=None, kill_by=None, record_active_gen_sizes=False):
        """
        Main method for running a single realization of the epidemic simulation.

        :param with_memberships: Set to True if specified memberships in Simulation configuration.
        :param uniform_rate: Will use a uniform transmission and recovery event probabilites instead of a customized matrix. Increases simulation speed, use if possible.
        :param wait_for_recovery: Will run simulation until all possible nodes have recovered.
        """
        self.use_uniform_rate = uniform_rate
        if self.use_uniform_rate and (self._uniform_beta_es is None or self._uniform_gamma_ei is None or
                                      self._uniform_gamma is None or self._uniform_beta is None):
            raise Exception('Since uniform_rate=True, must set all transmission and recovery parameter values via '
                            'set_uniform_{param} method}')
        if with_memberships:
            self.track_memberships = True
        if self.track_memberships:
            self._init_membership_state_time_series()
        self._initialize_patient_zero(label=p_zero)
        self._gens_size_over_time = []
        self._gens_size_over_time.append(list(self._current_active_gen_sizes))
        while self._current_sim_time < self.total_sim_time:
            # Run one step
            self._single_step(uniform_rate=uniform_rate, visualize=visualize, viz_graph=viz_graph, viz_pos=viz_pos,
                              record_active_gen_sizes=record_active_gen_sizes)

            self._total_num_timesteps += 1
            if self._current_number_IS_events == 0 and self._current_number_ES_events == 0:
                if not wait_for_recovery:
                    break
                elif len(self._recovery_events) == 0:
                    break
            if kill_by is not None and self._highest_gen >= kill_by:
                should_quit = self._check_active_gens(kill_by)
                if should_quit:
                    break

    def _single_step(self, visualize=False, uniform_rate=True, viz_graph=None, viz_pos=None,
                     record_active_gen_sizes=False):
        self.use_uniform_rate = uniform_rate
        if visualize:
            self._visualize_network(viz_graph, viz_pos)
        event_class, next_event, tau = self.draw_event_class()
        self._current_sim_time += tau
        self._time_series.append(self._time_series[-1] + tau)
        self._real_time_srs_infc.append(
            self._current_number_recovery_events)
        self._real_time_srs_exp.append(self._current_number_EI_events)
        self._real_time_srs_rec.append(len(self._recovered_nodes))
        if record_active_gen_sizes:
            self._gens_size_over_time.append(list(self._current_active_gen_sizes))
        if self.track_memberships:
            self._record_membership_states()

        if event_class is not None:
            if event_class == eventtype.EventType.INFECTEDSUSCEPTIBLE:
                infection_event = next_event
                infection_event.expose()
                self._exposed_to_infected_events[infection_event.get_right_node().get_label()] = infection_event.get_right_node()
                self._exposed_to_infected_events_keys.append(infection_event.get_right_node().get_label())
                self._current_number_EI_events += 1
                self._max_number_EI_events += 1

                try:  # If they are a member of an existing generation, bookkeeping will happen here:
                    self._gen_collection[infection_event.get_right_node().get_generation()].append(
                        infection_event.get_right_node().get_label())
                    self._gen_collection_active[infection_event.get_right_node().get_generation()].append(
                        infection_event.get_right_node().get_label())
                    self.active_gen_ts.append(self.active_gen_ts[-1])  # Active gens stays the same this round
                    self.total_gen_ts.append(self.total_gen_ts[-1])  # Total gens stays the same this round
                    self._current_active_gen_sizes[
                        infection_event.get_right_node().get_generation()] += 1  # One more active member of the generation
                except KeyError:  # If they are the first member of a new generation, book keeping happens here
                    self._gen_collection[infection_event.get_right_node().get_generation()] = [
                        infection_event.get_right_node().get_label()]
                    self._gen_collection_active[infection_event.get_right_node().get_generation()] = [
                        infection_event.get_right_node().get_label()]
                    self._highest_gen += 1
                    self._generational_emergence[self._highest_gen] = self._current_sim_time
                    self.active_gen_ts.append(self.active_gen_ts[-1] + 1)  # Active gens increases by one
                    self.total_gen_ts.append(self.total_gen_ts[-1] + 1)  # Total gens increases by one
                    self._current_active_gen_sizes[
                        infection_event.get_right_node().get_generation()] += 1
                self._update_IS_events(infection_IS_event=infection_event)
                self._update_ES_events(infection_ES_event=infection_event)
                self._add_ES_events(infection_event.get_right_node())
            elif event_class == eventtype.EventType.EXPOSEDSUSCEPTIBLE:
                exposure_event = next_event
                exposure_event.expose()
                self._exposed_to_infected_events[exposure_event.get_right_node().get_label()] = exposure_event.get_right_node()
                self._exposed_to_infected_events_keys.append(exposure_event.get_right_node().get_label())
                self._current_number_EI_events += 1
                self._max_number_EI_events += 1
                try:  # If they are a member of an existing generation, bookkeeping will happen here:
                    self._gen_collection[exposure_event.get_right_node().get_generation()].append(
                        exposure_event.get_right_node().get_label())
                    self._gen_collection_active[exposure_event.get_right_node().get_generation()].append(
                        exposure_event.get_right_node().get_label())
                    self.active_gen_ts.append(self.active_gen_ts[-1])  # Active gens stays the same this round
                    self.total_gen_ts.append(self.total_gen_ts[-1])  # Total gens stays the same this round
                    self._current_active_gen_sizes[
                        exposure_event.get_right_node().get_generation()] += 1  # One more active member of the generation
                except KeyError:  # If they are the first member of a new generation, book keeping happens here
                    self._gen_collection[exposure_event.get_right_node().get_generation()] = [
                        exposure_event.get_right_node().get_label()]
                    self._gen_collection_active[exposure_event.get_right_node().get_generation()] = [
                        exposure_event.get_right_node().get_label()]
                    self._highest_gen += 1
                    self._generational_emergence[self._highest_gen] = self._current_sim_time
                    self.active_gen_ts.append(self.active_gen_ts[-1] + 1)  # Active gens increases by one
                    self.total_gen_ts.append(self.total_gen_ts[-1] + 1)  # Total gens increases by one
                    self._current_active_gen_sizes[
                        exposure_event.get_right_node().get_generation()] += 1
                self._update_IS_events(infection_IS_event=exposure_event)
                self._update_ES_events(infection_ES_event=exposure_event)
                self._add_ES_events(exposure_event.get_right_node())

            elif event_class == eventtype.EventType.EXPOSEDINFECTED:
                exposed_infected_event = next_event
                exposed_infected_event.infect()
                self._current_number_recovery_events += 1
                self._max_num_recovery_events += 1
                self._recovery_events[exposed_infected_event.get_label()] = exposed_infected_event
                self._recovery_events_keys.append(exposed_infected_event.get_label())

                self._exposed_to_infected_events.pop(exposed_infected_event.get_label())
                self._current_number_EI_events -= 1

                self._update_ES_events(exposed_infected_event=exposed_infected_event)
                self._add_IS_events(exposed_infected_event)

            elif event_class == eventtype.EventType.RECOVER:
                recovery_event = next_event
                recovery_event.recover()
                self._recovery_events.pop(recovery_event.get_label())
                self._current_number_recovery_events -= 1
                self._update_IS_events(recovery_event=recovery_event)
                self._recovered_nodes.append(recovery_event)
                self._current_active_gen_sizes[
                    recovery_event.get_generation()] -= 1  # One less active member of the generation
                try:
                    self._gen_collection_active[recovery_event.get_generation()].remove(recovery_event.get_label())
                except ValueError:
                    pass
                if len(self._gen_collection_active[
                           recovery_event.get_generation()]) == 0:  # If this was the last remaining member of the generation
                    if self.active_gen_ts[-1] == 0:
                        self.active_gen_ts.append(
                            self.active_gen_ts[-1])  # If active generations is already 0, then stays at 0
                    else:
                        self.active_gen_ts.append(
                            self.active_gen_ts[-1] - 1)  # Otherwise, reduces active generations by 1
                else:
                    self.active_gen_ts.append(self.active_gen_ts[-1])
                self.total_gen_ts.append(self.total_gen_ts[-1])  # this always should stay the same no matter what

        self._current_sim_time += tau

    def _update_ES_events(self, infection_ES_event=None, exposed_infected_event=None):
        if infection_ES_event is not None:
            try:
                in_degree_events = self._in_degree_ES_events[infection_ES_event.get_right_node().get_label()]
                for event in in_degree_events:
                    left_node_idx = event.get_left_node().get_label()
                    try:
                        self._out_degree_ES_events[left_node_idx].remove(event)
                        self._out_degree_ES_lengths[left_node_idx] -= 1
                        self._current_number_ES_events -= 1
                        flat_key = self._edge_keys_es[event.get_left_node().get_label()][
                            event.get_right_node().get_label()]
                        self._potential_ES_events_flat[flat_key] = None
                    except ValueError:
                        pass
                self._in_degree_ES_events[infection_ES_event.get_right_node().get_label()] = []
            except KeyError:
                pass
        elif exposed_infected_event is not None:
            try:
                out_degree_events = self._out_degree_ES_events[exposed_infected_event.get_label()]
                for event in out_degree_events:
                    flat_key = self._edge_keys_es[event.get_left_node().get_label()][event.get_right_node().get_label()]
                    try:
                        self._potential_ES_events_flat[flat_key] = None
                    except:
                        print(flat_key)
                self._out_degree_ES_events[exposed_infected_event.get_label()] = []
                self._out_degree_ES_lengths[exposed_infected_event.get_label()] = 0
                self._current_number_ES_events -= len(out_degree_events)
            except KeyError:
                pass
            try:
                self._in_degree_ES_events[exposed_infected_event.get_label()] = []
            except KeyError:
                pass

    def _add_ES_events(self, infected_node):
        infection_adjlist = self._adjlist[infected_node.get_label()]
        adjlist_length = len(infection_adjlist)
        if adjlist_length > 1:
            length = adjlist_length
            for n in range(1, length):
                j = infection_adjlist[n]
                if self.use_uniform_rate:
                    candidate_node = network.Node(j, -1, nodestate.NodeState.SUSCEPTIBLE, self._uniform_gamma_ei)
                else:
                    candidate_node = network.Node(j, -1, nodestate.NodeState.SUSCEPTIBLE,
                                                  self._Theta_ExposedInfected[j])
                if self.track_memberships:
                    candidate_node.set_membership(self._node_memberships[candidate_node.get_label()])
                neighbor_node = self._existing_node(candidate_node)  # 18 seconds
                if neighbor_node.get_state() == nodestate.NodeState.SUSCEPTIBLE:
                    if self.use_uniform_rate:
                        edge_ij = network.Edge(infected_node, neighbor_node, self._uniform_beta)
                    else:
                        edge_ij = network.Edge(infected_node, neighbor_node,
                                               self._Beta_ExposedSusceptible[infected_node.get_label()][j])
                    try:
                        self._out_degree_ES_events[infected_node.get_label()].append(edge_ij)
                        self._out_degree_ES_lengths[infected_node.get_label()] += 1
                    except KeyError:
                        self._out_degree_ES_events[infected_node.get_label()] = [edge_ij]
                        self._out_degree_ES_lengths[infected_node.get_label()] = 1
                    try:
                        self._in_degree_ES_events[neighbor_node.get_label()].append(edge_ij)
                    except KeyError:
                        self._in_degree_ES_events[neighbor_node.get_label()] = [edge_ij]
                    self._current_number_ES_events += 1

                    self._potential_ES_events_flat.append(edge_ij)
                    try:
                        self._edge_keys_es[edge_ij.get_left_node().get_label()][
                            edge_ij.get_right_node().get_label()] = self._next_key_es
                    except KeyError:
                        self._edge_keys_es[edge_ij.get_left_node().get_label()] = {
                            edge_ij.get_right_node().get_label(): self._next_key_es}
                    self._next_key_es += 1

    def _record_membership_states(self):
        # TODO assign a time series vector for number of groups for membership
        # ex. current_infected_group2.append(len(current_infected_nodes where membership==group2)
        for group in self._membership_groups:
            infected_in_group = list(node for node in self._active_node_dict.values() if
                                     node.get_state() == nodestate.NodeState.INFECTED and node.get_membership() == group)
            exposed_in_group = list(node for node in self._active_node_dict.values() if
                                     node.get_state() == nodestate.NodeState.EXPOSED and node.get_membership() == group)
            recovered_in_group = list(node for node in self._active_node_dict.values() if
                                     node.get_state() == nodestate.NodeState.RECOVERED and node.get_membership() == group)
            self._membership_time_series_infc[group].append(len(infected_in_group))
            self._membership_time_series_exp[group].append(len(exposed_in_group))
            self._membership_time_series_rec[group].append(len(recovered_in_group))

    def _init_membership_state_time_series(self):
        self.membership_time_series_infc = {}
        for group in self._membership_groups:
            self._membership_time_series_infc[group] = []
            self._membership_time_series_rec[group] = []
            self._membership_time_series_exp[group] = []

    def tabulate_continuous_time(self, time_buckets=100, custom_range=False, custom_t_lim=100, active_gen_info=False,
                                 active_gen_sizes=False):
        max_time = max(self._time_series)
        time_partition = np.linspace(0, max_time + 1, time_buckets, endpoint=False)
        if custom_range:
            time_partition = np.linspace(0, custom_t_lim, time_buckets, endpoint=False)
        exposed_time_series = np.zeros(len(time_partition))
        infection_time_series = np.zeros(len(time_partition))
        recover_time_series = np.zeros(len(time_partition))
        ts_length = len(self._time_series)

        active_gen_time_series = np.zeros(len(time_partition))
        total_gen_time_series = np.zeros(len(time_partition))
        active_gen_sizes_ts = np.zeros((len(time_partition), 100))
        if active_gen_sizes:
            self._gens_size_over_time = np.array(self._gens_size_over_time)

        idx_floor = 0
        for t in range(ts_length - 1):
            real_time_val = self._time_series[t]
            real_time_infected = self._real_time_srs_infc[t]
            real_time_exposed = self._real_time_srs_exp[t]
            real_time_recovered = self._real_time_srs_rec[t]
            if active_gen_info:
                real_time_active_gens = self.active_gen_ts[t]
                real_time_total_gens = self.total_gen_ts[t]
            if active_gen_sizes:
                real_time_gen_sizes = self._gens_size_over_time[t]
            found_soonest = False
            idx = idx_floor
            while not found_soonest and idx<time_buckets:
                upper_val = time_partition[idx]
                if real_time_val <= upper_val:
                    idx_floor = idx
                    infection_time_series[idx:] = real_time_infected
                    recover_time_series[idx:] = real_time_recovered
                    exposed_time_series[idx:] = real_time_exposed
                    if active_gen_info:
                        active_gen_time_series[idx:] = real_time_active_gens
                        total_gen_time_series[idx:] = real_time_total_gens
                    if active_gen_sizes:
                        active_gen_sizes_ts[idx:] = real_time_gen_sizes
                    found_soonest = True
                idx += 1
        if active_gen_info and active_gen_sizes:
            return time_partition, infection_time_series, recover_time_series, exposed_time_series, \
                   active_gen_time_series, total_gen_time_series, active_gen_sizes_ts
        elif active_gen_info:
            return time_partition, infection_time_series, recover_time_series, exposed_time_series,\
                   active_gen_time_series, total_gen_time_series
        return time_partition, infection_time_series, recover_time_series, exposed_time_series

    def tabulate_continuous_time_with_groups(self, time_buckets=100, custom_range=False, custom_t_lim=100):
        max_time = max(self._time_series)
        time_partition = np.linspace(0, max_time + 1, time_buckets, endpoint=False)
        if custom_range:
            time_partition = np.linspace(0, custom_t_lim, time_buckets, endpoint=False)
        infection_time_series_group_dict = {}
        exposed_time_series_group_dict = {}
        recovered_time_series_group_dict = {}
        # infection_time_series = np.zeros(len(time_partition))
        for group in self._membership_groups:
            infection_time_series_group_dict[group] = np.zeros(len(time_partition))
            exposed_time_series_group_dict[group] = np.zeros(len(time_partition))
            recovered_time_series_group_dict[group] = np.zeros(len(time_partition))

        ts_length = len(self._time_series)

        idx_floor = 0
        for t in range(ts_length - 1):
            real_time_val = self._time_series[t]
            found_soonest = False
            idx = idx_floor
            while not found_soonest and idx<time_buckets:
                upper_val = time_partition[idx]
                if real_time_val <= upper_val:
                    idx_floor = idx
                    for group in self._membership_groups:
                        real_time_infected = self._membership_time_series_infc[group][t]
                        real_time_exposed = self._membership_time_series_exp[group][t]
                        real_time_recovered = self._membership_time_series_rec[group][t]
                        infection_time_series_group_dict[group][idx:] = real_time_infected
                        exposed_time_series_group_dict[group][idx:] = real_time_exposed
                        recovered_time_series_group_dict[group][idx:] = real_time_recovered
                    found_soonest = True
                idx += 1

        return time_partition, infection_time_series_group_dict, exposed_time_series_group_dict, recovered_time_series_group_dict

    def draw_event_class(self):
        partition_end_markers = {}
        num_of_recovery_events = self._current_number_recovery_events
        num_of_expose_possible_events = self._current_number_IS_events
        num_of_infect_possible_events = self._current_number_ES_events
        num_of_expose_to_infection_events = self._current_number_EI_events
        partition_end_markers[0] = num_of_expose_possible_events * self._uniform_beta
        partition_end_markers[1] = num_of_infect_possible_events * self._uniform_beta_es + partition_end_markers[0]
        partition_end_markers[2] = num_of_expose_to_infection_events * self._uniform_gamma_ei + partition_end_markers[1]
        partition_end_markers[3] = num_of_recovery_events * self._uniform_gamma + partition_end_markers[2]
        random_draw = random.uniform(0, partition_end_markers[3])
        sum_of_rates = (num_of_expose_possible_events * self._uniform_beta +
                        num_of_infect_possible_events * self._uniform_beta_es +
                        num_of_expose_to_infection_events * self._uniform_gamma_ei +
                        num_of_recovery_events * self._uniform_gamma)
        tau = random.expovariate(max(sum_of_rates, .0000001))
        if random_draw < partition_end_markers[0]:
            found_next_event = None
            while found_next_event is None:
                random_idx = int(random.uniform(0, self._next_key))
                found_next_event = self._potential_IS_events_flat[random_idx]
            return eventtype.EventType.INFECTEDSUSCEPTIBLE, found_next_event, tau
        elif random_draw < partition_end_markers[1]:
            found_next_event = None
            while found_next_event is None:
                random_idx = int(random.uniform(0, self._next_key_es))
                found_next_event = self._potential_ES_events_flat[random_idx]
            return eventtype.EventType.EXPOSEDSUSCEPTIBLE, found_next_event, tau
        elif random_draw < partition_end_markers[2]:
            found_next_event = None
            while found_next_event is None:
                random_idx = random.randint(0, self._max_number_EI_events)
                try:
                    next_event_label = self._exposed_to_infected_events_keys[random_idx]
                except IndexError:
                    continue
                try:
                    found_next_event = self._exposed_to_infected_events[next_event_label]
                except KeyError:
                    continue
            return eventtype.EventType.EXPOSEDINFECTED, found_next_event, tau
        else:
            found_next_event = None
            while found_next_event is None:
                try:
                    random_idx = random.randint(0, self._max_num_recovery_events)
                except ValueError:
                    print('Value error')
                    print(f'number recovery nodes {print(len(self._recovery_events))}')
                next_event_label = self._recovery_events_keys[random_idx]
                try:
                    found_next_event = self._recovery_events[next_event_label]
                except KeyError:
                    continue
            return eventtype.EventType.RECOVER, found_next_event, tau



import unittest
import numpy as np
from epintervene.simobjects import simulation
import networkx as nx
import matplotlib.pyplot as plt

#TODO needs to be edited

class TestSimulation(unittest.TestCase):
    def test_sim_creates_patientzero(self):
        A = np.random.random_integers(0, 1, (10, 10))
        A = (A + A.T)/2
        sim = simulation.Simulation(A)
        self.assertRaises(AttributeError, sim._initialize_patient_zero)

        good_Beta = np.full((10,10), 0.2)
        good_Gamma = np.full(10, 0.5)
        sim.add_recover_event_rates(good_Gamma)
        sim.add_infection_event_rates(good_Beta)

        sim._initialize_patient_zero()
        self.assertEqual(len(sim._potential_recovery_events._event_list), 1)

    def test_beta_rate_matrix_throws_exception(self):
        A = np.random.random_integers(0, 1, (10, 10))
        A = (A + A.T)/2
        sim = simulation.Simulation(A)
        bad_Beta = np.zeros((2,2))
        self.assertRaises(ValueError, sim.add_infection_event_rates, bad_Beta)

    def test_beta_rate_matrix_alters_event_rates(self):
        A = np.random.random_integers(0, 1, (10, 10))
        A = (A + A.T)/2
        sim = simulation.Simulation(A)
        good_Beta = np.full((10,10), 0.2)
        good_Gamma = np.full(10, 0.5)
        for edge in sim._potential_IS_events._event_list:
            self.assertEqual(edge.event_rate, 0)

        sim.add_infection_event_rates(good_Beta)
        sim.add_recover_event_rates(good_Gamma)
        for edge in sim._potential_IS_events._event_list:
            self.assertEqual(edge.event_rate, 0.2)

    def test_IS_edges_are_updated_after_single_step(self):
        N = 10
        self.beta = 0.99
        self.gamma = 0.0001
        graph = nx.generators.complete_graph(N)
        N = len(graph.nodes())
        self.Beta = np.full((N, N), self.beta)
        self.Gamma = np.full(N, self.gamma)
        adjacency_matrix = np.array(nx.adjacency_matrix(graph).todense())
        sim = simulation.Simulation(adjacency_matrix)
        sim.add_infection_event_rates(self.Beta)
        sim.add_recover_event_rates(self.Gamma)
        sim._initialize_patient_zero()
        print('before single step')

        self.assertGreaterEqual(len(sim._potential_IS_events._event_list), 1)
        self.assertEqual(len(sim._potential_recovery_events._event_list), 1)
        sim._single_step()

        print('after single step')
        self.assertEqual(len(sim._potential_recovery_events._event_list), 2)
        infected_nodes = list(map(lambda node: node.get_label(), sim._potential_recovery_events._event_list))
        for edge in sim._potential_IS_events._event_list:
            self.assertIn(edge.get_left_node().get_label(), infected_nodes)
            self.assertNotIn(edge.get_right_node().get_label(), infected_nodes)
        if len(sim._potential_IS_events._event_list) > 0:
            sim._single_step()
            print('after second single step')
            self.assertEqual(len(sim._potential_recovery_events._event_list), 3)
            infected_nodes = list(map(lambda node: node.get_label(), sim._potential_recovery_events._event_list))
            for edge in sim._potential_IS_events._event_list:
                self.assertIn(edge.get_left_node().get_label(), infected_nodes)
                self.assertNotIn(edge.get_right_node().get_label(), infected_nodes)

    def test_custom_time_series_results(self):
        N = 100
        self.beta = 0.99
        self.gamma = 0.0001
        graph = nx.generators.erdos_renyi_graph(N, 0.02)
        N = len(graph.nodes())
        self.Beta = np.full((N, N), self.beta)
        self.Gamma = np.full(N, self.gamma)
        adjacency_matrix = np.array(nx.adjacency_matrix(graph).todense())
        sim = simulation.Simulation(adjacency_matrix)
        sim.add_infection_event_rates(self.Beta)
        sim.add_recover_event_rates(self.Gamma)
        sim._initialize_patient_zero()

        sim.run_sim()

        custom_time_results = sim.tabulate_continuous_time(time_buckets=1000, custom_range=True, custom_t_lim=15)
        self.assertEqual(len(custom_time_results[0]), 1000)
        self.assertEqual(max(custom_time_results[0]), 15-0.015)

if __name__ == '__main__':
    unittest.main()



from epintervene.simobjects.simulation import Simulation, SimulationSEIR
from epintervene.simobjects.extended_simulation import RandomRolloutSimulation, RandomInterventionSim, UniversalInterventionSim, AbsoluteTimeNetworkSwitchSim, TargetedInterventionSim, TargetedRolloutSimulation
from epintervene.simobjects.network import NetworkBuilder, Node, Edge


import unittest
import numpy as np
from epintervene.simobjects import simulation
from epintervene.simobjects import extended_simulation
import networkx as nx
import matplotlib.pyplot as plt


class TestRandomRolloutSimulation(unittest.TestCase):
    def test_vaccinated_nodes_are_not_done_twice(self):
        N = 10
        self.beta = 0.99
        self.gamma = 0.0001
        graph = nx.generators.complete_graph(N)
        N = len(graph.nodes())
        self.Beta = np.full((N, N), self.beta)
        self.Gamma = np.full(N, self.gamma)
        adjacency_matrix = np.array(nx.adjacency_matrix(graph).todense())
        sim = extended_simulation.RandomRolloutSimulation(adjacency_matrix)
        sim.add_infection_event_rates(self.Beta)
        sim.add_recover_event_rates(self.Gamma)
        sim.configure_intervention([1, 2], [0.0, 0.0], [.9, .1])

        first_round_vaccinated_nodes = sim.intervene(0)
        node_labels = [node._label for node in first_round_vaccinated_nodes]
        first_round_unique_nodes = np.unique(node_labels)
        self.assertEqual(len(first_round_unique_nodes), 9)

        second_round_vaccinated_nodes = sim.intervene(1)
        second_round_node_labels = [node._label for node in second_round_vaccinated_nodes]
        second_round_unique_nodes = np.unique(second_round_node_labels)
        self.assertEqual(len(second_round_unique_nodes), 1)
        for nl in second_round_unique_nodes:
            self.assertNotIn(nl, first_round_unique_nodes)



if __name__ == '__main__':
    unittest.main()


import unittest
from epintervene.simobjects import network
from epintervene.simobjects import nodestate

#TODO: needs to be edited

class TestNetwork(unittest.TestCase):
    def test_node(self):
        first_node = network.Node(label=1, generation=0, state=nodestate.NodeState.SUSCEPTIBLE, event_rate=.1)
        second_node = network.Node(label=2, generation=1, state=nodestate.NodeState.INFECTED, event_rate=.2)
        self.assertFalse(first_node.equals(second_node))

        first_node.infect()
        self.assertEqual(first_node._state, nodestate.NodeState.INFECTED)
        first_node.recover()
        self.assertEqual(first_node._state, nodestate.NodeState.RECOVERED)

        third_node_same_label = network.Node(label=1, generation=0, state=nodestate.NodeState.SUSCEPTIBLE, event_rate=0.1)
        self.assertTrue(third_node_same_label.equals(first_node))

    def test_edge(self):
        first_node = network.Node(label=1, generation=0, state=nodestate.NodeState.SUSCEPTIBLE, event_rate=.1)
        second_node = network.Node(label=2, generation=1, state=nodestate.NodeState.INFECTED, event_rate=.2)

        edge = network.Edge(first_node, second_node, event_rate=0.5)
        should_be_same_edge = network.Edge(first_node, second_node, event_rate=0.4)

        self.assertTrue(edge.equals(should_be_same_edge))

    def test_edge_infection(self):
        first_node = network.Node(label=1, generation=0, state=nodestate.NodeState.INFECTED, event_rate=.1)
        second_node = network.Node(label=2, generation=-1, state=nodestate.NodeState.SUSCEPTIBLE, event_rate=.2)

        self.assertEqual(first_node.get_state(), nodestate.NodeState.INFECTED)
        self.assertEqual(first_node.get_generation(), 0)
        self.assertEqual(second_node.get_state(), nodestate.NodeState.SUSCEPTIBLE)
        self.assertEqual(second_node.get_generation(), -1)

        edge = network.Edge(first_node, second_node, event_rate=0.5)

        edge.infect()

        self.assertEqual(second_node.get_state(), nodestate.NodeState.INFECTED)
        self.assertEqual(second_node.get_generation(), 1)


if __name__ == '__main__':
    unittest.main()



from setuptools import find_packages
import setuptools

VERSION = '0.1.0'
DESCRIPTION = 'EpIntervene Simulation Package'
LONG_DESCRIPTION = 'EpIntervene package for simulating SIR and SEIR Epidemics with or without interventions'

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="epintervene",
    version=VERSION,
    author="Andrea Allen",
    author_email="andrea2allen@gmail.com",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=[],  # add any additional packages that
    # needs to be installed along with your package.

    keywords=['python', 'epintervene'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
    python_requires=">=3.6",
)


from epintervene.simobjects import simulation
import numpy as np
from epintervene.simobjects import network
from epintervene.simobjects import nodestate
import random


class UniversalInterventionSim(simulation.Simulation):
    def __init__(self, N, adjlist=None, max_unitless_sim_time=1000000, membership_groups=None, node_memberships=None):
        super().__init__(N=N, adj_list=adjlist, max_unitless_sim_time=max_unitless_sim_time,
                         membership_groups=membership_groups,
                         node_memberships=node_memberships)
        self.intervention_gen = None
        self.beta_redux = None
        self.intervened = False
        self.time_of_intervention = max_unitless_sim_time
        self.use_uniform_rate = True

    def simtype(self):
        print('I am a simulation class of type universal intervention')

    def configure_intervention(self, intervention_gen, beta_redux):
        self.intervention_gen = intervention_gen
        self.beta_redux = beta_redux

    def run_sim(self, with_memberships=False, uniform_rate=True, wait_for_recovery=False, visualize=False,
                viz_graph=None, viz_pos=None, p_zero=None, kill_by=None, record_active_gen_sizes=False):
        self.use_uniform_rate = uniform_rate
        if self.use_uniform_rate and (
                self._uniform_beta is None or self._uniform_gamma is None):
            raise Exception(
                'Since uniform_rate=True, must set all transmission and recovery parameter values via set_uniform_{param} method}')
        if with_memberships:
            self.track_memberships = True
        if self.track_memberships:
            self._init_membership_state_time_series()
        self._initialize_patient_zero(label=p_zero)
        self._gens_size_over_time = []
        self._gens_size_over_time.append(list(self._current_active_gen_sizes))
        while self._current_sim_time < self.total_sim_time:
            if not self.intervened:
                if self._highest_gen >= self.intervention_gen:
                    self.intervene()
                    self.time_of_intervention = self._current_sim_time
                    self.intervened = True
            # Run one step
            self._single_step(uniform_rate=uniform_rate, visualize=visualize, viz_graph=viz_graph, viz_pos=viz_pos,
                              record_active_gen_sizes=record_active_gen_sizes)

            self._total_num_timesteps += 1
            if self._current_number_IS_events == 0:
                if not wait_for_recovery:
                    break
                elif len(self._recovery_events) == 0:
                    break
            if kill_by is not None and self._highest_gen >= kill_by:
                should_quit = self._check_active_gens(kill_by)
                if should_quit:
                    break

    def intervene(self, reduce_current_edges=False):
        print('intervening')
        N = self._N
        if self.use_uniform_rate:
            self._uniform_beta = self.beta_redux
        else:
            new_Beta = np.full((N, N), self.beta_redux)
            self.Beta = new_Beta
        # # change event rate for each existing edge pair
        # # NO LONGER NECESSARY WITH REQUIRED UNIFORM RATE
        # if reduce_current_edges:
        #     for edge in self._potential_IS_events._event_list:
        #         if self.use_uniform_rate:
        #             edge.set_event_rate(self.beta_redux)
        #         else:
        #             edge.set_event_rate(self.Beta[edge.i.label][edge.j.label])


class RandomInterventionSim(simulation.Simulation):
    def __init__(self, N, adjlist=None, max_unitless_sim_time=1000000, membership_groups=None, node_memberships=None):
        super().__init__(N=N, adj_list=adjlist, max_unitless_sim_time=max_unitless_sim_time,
                         membership_groups=membership_groups,
                         node_memberships=node_memberships)
        self.intervention_gen = None
        self.beta_redux = None
        self.proportion_reduced = None
        self.intervened = False
        self.time_of_intervention = max_unitless_sim_time
        self.use_uniform_rate = True

    def simtype(self):
        print('I am a simulation class of type Random Intervention')

    def configure_intervention(self, intervention_gen, beta_redux, proportion_reduced):
        self.intervention_gen = intervention_gen
        self.beta_redux = beta_redux
        self.proportion_reduced = proportion_reduced

    def run_sim(self, with_memberships=False, uniform_rate=True, wait_for_recovery=False, visualize=False,
                viz_graph=None, viz_pos=None, p_zero=None, kill_by=None, record_active_gen_sizes=False):
        self.use_uniform_rate = uniform_rate
        if with_memberships: self.track_memberships = True
        self.use_uniform_rate = uniform_rate
        if self.track_memberships:
            self._init_membership_state_time_series()
        self._initialize_patient_zero()
        while self._current_sim_time < self.total_sim_time:
            if not self.intervened:
                if self._highest_gen >= self.intervention_gen:
                    self.intervene()
                    self.time_of_intervention = self._current_sim_time
                    self.intervened = True
            # Run one step
            self._single_step(uniform_rate=uniform_rate, visualize=visualize, viz_graph=viz_graph, viz_pos=viz_pos,
                              record_active_gen_sizes=record_active_gen_sizes)

            self._total_num_timesteps += 1
            if self._current_number_IS_events == 0:
                if not wait_for_recovery:
                    break
                elif len(self._recovery_events) == 0:
                    break
            if kill_by is not None and self._highest_gen >= kill_by:
                should_quit = self._check_active_gens(kill_by)
                if should_quit:
                    break

    def intervene(self):
        print('intervening')
        if self._N == 0:
            self._N = len(self._adjlist)
        vaccinated_nodes = []
        vax_labels = []
        random_set = list(np.where(np.random.uniform(0, 1, self._N) <= self.proportion_reduced)[0])
        for node_label in random_set:
            if self.use_uniform_rate:
                candidate_node = network.Node(node_label, -1, None, self._uniform_gamma)
            else:
                candidate_node = network.Node(node_label, -1, None, self._Gamma[node_label])
            if self.track_memberships:
                candidate_node.set_membership(self._node_memberships[candidate_node.get_label()])
            existing_node = self._existing_node(candidate_node)
            if existing_node.get_state() != nodestate.NodeState.VACCINATED:
                if existing_node.get_state() != nodestate.NodeState.INFECTED:
                    existing_node.vaccinate()
                    vaccinated_nodes.append(existing_node)
                    vax_labels.append(node_label)
                # self._Beta[node_label] = np.full(N, self.beta_redux)
                # self._Beta[:, node_label] = np.full(N, self.beta_redux).T
                    self._update_IS_events(recovery_event=existing_node)


class RandomRolloutSimulation(simulation.Simulation):
    def __init__(self, N, adjlist=None, max_unitless_sim_time=1000000, membership_groups=None, node_memberships=None):
        super().__init__(N=N, adj_list=adjlist, max_unitless_sim_time=max_unitless_sim_time,
                         membership_groups=membership_groups,
                         node_memberships=node_memberships)
        self.intervention_gen_list = None
        self.beta_redux_list = None
        self.proportion_reduced_list = None
        self.intervened_status_list = []
        self.time_of_intervention_list = []
        self.next_up_intervention_entry = 0
        self.use_uniform_rate = True

    def simtype(self):
        print('I am a simulation class of type Multi intervention Random Intervention')

    def configure_intervention(self, intervention_gen_list, beta_redux_list, proportion_reduced_list):
        self.intervention_gen_list = intervention_gen_list
        self.beta_redux_list = beta_redux_list
        self.proportion_reduced_list = proportion_reduced_list
        for i in range(len(intervention_gen_list)):
            self.intervened_status_list.append(False)

    def run_sim(self, with_memberships=False, uniform_rate=True, wait_for_recovery=False, visualize=False,
                viz_graph=None, viz_pos=None, p_zero=None, kill_by=None, record_active_gen_sizes=False):
        self.use_uniform_rate = uniform_rate
        if with_memberships: self.track_memberships = True
        self.use_uniform_rate = uniform_rate
        if self.track_memberships:
            self._init_membership_state_time_series()
        self._initialize_patient_zero()
        while self._current_sim_time < self.total_sim_time:
            if self.next_up_intervention_entry < len(self.intervention_gen_list):
                if not self.intervened_status_list[self.next_up_intervention_entry]:
                    if self._highest_gen >= self.intervention_gen_list[self.next_up_intervention_entry]:
                        self.intervene(self.next_up_intervention_entry)
                        self.time_of_intervention_list.append(self._current_sim_time)
                        self.intervened_status_list[self.next_up_intervention_entry] = True
                        self.next_up_intervention_entry += 1
            # Run one step
            self._single_step(uniform_rate=uniform_rate, visualize=visualize, viz_graph=viz_graph, viz_pos=viz_pos,
                              record_active_gen_sizes=record_active_gen_sizes)

            self._total_num_timesteps += 1
            if self._current_number_IS_events == 0:
                if not wait_for_recovery:
                    break
                elif len(self._recovery_events) == 0:
                    break
            if kill_by is not None and self._highest_gen >= kill_by:
                should_quit = self._check_active_gens(kill_by)
                if should_quit:
                    break

    def intervene(self, intervention_entry):
        print('intervening')
        if self._N == 0:
            self._N = len(self._A[0])
        frac_of_network = self.proportion_reduced_list[intervention_entry] * self._N
        how_many = 0
        if frac_of_network > 1:
            how_many = int(np.round(frac_of_network, 0))
        vaccinated_nodes = []
        vax_labels = []
        while len(vaccinated_nodes) < how_many:
            random_set = np.unique(np.random.randint(0, self._N, how_many)) #TODO need to fix random set here like above
            for node_label in random_set:
                if len(vax_labels) < how_many:
                    if node_label not in vax_labels:
                        if self.use_uniform_rate:
                            candidate_node = network.Node(node_label, -1, None, self._uniform_gamma)
                        else:
                            candidate_node = network.Node(node_label, -1, None, self._Gamma[node_label])
                        if self.track_memberships:
                            candidate_node.set_membership(self._node_memberships[candidate_node.get_label()])
                        existing_node = self._existing_node(candidate_node)
                        # all the nodes become active
                        if existing_node.get_state() != nodestate.NodeState.VACCINATED:
                            # these should also be added to the active nodes
                            if existing_node.get_state() != nodestate.NodeState.INFECTED:
                                existing_node.vaccinate()
                                vaccinated_nodes.append(existing_node)
                                vax_labels.append(node_label)
                            # self._Beta[node_label] = np.full(self._N, self.beta_redux_list[intervention_entry])
                            # self._Beta[:, node_label] = np.full(self._N, self.beta_redux_list[intervention_entry]).T
                                self._update_IS_events(recovery_event=existing_node)
                            # else:
                            #     print("I was already infected")
        return vaccinated_nodes


class TargetedInterventionSim(simulation.Simulation):
    def __init__(self, N, adjlist=None, max_unitless_sim_time=1000000, membership_groups=None, node_memberships=None):
        super().__init__(N=N, adj_list=adjlist, max_unitless_sim_time=max_unitless_sim_time,
                         membership_groups=membership_groups,
                         node_memberships=node_memberships)
        self.intervention_gen = None
        self.beta_redux = None
        self.proportion_reduced = None
        self.intervened = False
        self.time_of_intervention = max_unitless_sim_time
        self.use_uniform_rate = True

    def simtype(self):
        print('I am a simulation class of type Targeted Intervention')

    def configure_intervention(self, intervention_gen, beta_redux, proportion_reduced):
        self.intervention_gen = intervention_gen
        self.beta_redux = beta_redux
        self.proportion_reduced = proportion_reduced

    def run_sim(self, with_memberships=False, uniform_rate=True, wait_for_recovery=False, visualize=False,
                viz_graph=None, viz_pos=None, p_zero=None, kill_by=None, record_active_gen_sizes=False):
        self.use_uniform_rate = uniform_rate
        if with_memberships: self.track_memberships = True
        self.use_uniform_rate = uniform_rate
        if self.track_memberships:
            self._init_membership_state_time_series()
        self._initialize_patient_zero()
        while self._current_sim_time < self.total_sim_time:
            if not self.intervened:
                if self._highest_gen >= self.intervention_gen:
                    self.intervene()
                    self.time_of_intervention = self._current_sim_time
                    self.intervened = True
            # Run one step
            self._single_step(uniform_rate=uniform_rate, visualize=visualize, viz_graph=viz_graph, viz_pos=viz_pos,
                              record_active_gen_sizes=record_active_gen_sizes)

            self._total_num_timesteps += 1
            if self._current_number_IS_events == 0:
                if not wait_for_recovery:
                    break
                elif len(self._recovery_events) == 0:
                    break
            if kill_by is not None and self._highest_gen >= kill_by:
                should_quit = self._check_active_gens(kill_by)
                if should_quit:
                    break

    def intervene(self):
        print('intervening')
        if self._N == 0:
            self._N = len(self._adjlist)
        frac_of_network = self.proportion_reduced * self._N
        how_many = 0
        if frac_of_network > 1:
            how_many = int(np.round(frac_of_network, 0))

        degrees = list([len(neighbor) - 1 for neighbor in self._adjlist])
        degree_classes = {}
        for i in range(len(degrees)):
            degree = degrees[i]
            try:
                degree_classes[degree].append(i)
            except KeyError:
                degree_classes[degree] = [i]

        keys_decreasing = np.sort(list(degree_classes.keys()))[::-1]

        num_vaccinated = 0
        infct_hd_nodes = 0
        for degree in keys_decreasing:
            current_node_group = degree_classes[degree] # To match analytical prediction, once a degree class is bumped up to meet the quota, the whole class is vaccinated
            if num_vaccinated < how_many:
                for node in current_node_group:
                    if self.use_uniform_rate:
                        candidate_node = network.Node(node, -1, None, self._uniform_gamma)
                    else:
                        candidate_node = network.Node(node, -1, None, self._Gamma[node])
                    if self.track_memberships:
                        candidate_node.set_membership(self._node_memberships[candidate_node.get_label()])
                    existing_node = self._existing_node(candidate_node)
                    if existing_node.get_state() != nodestate.NodeState.VACCINATED:
                        if existing_node.get_state() != nodestate.NodeState.INFECTED: #Do not vaccinate currently infected nodes
                            existing_node.vaccinate()
                            num_vaccinated += 1 #Only counted if high degree nodes are successfully vaccinated
                        else:
                            infct_hd_nodes += 1 #Counts high-degree currently infected nodes
                            print(degree)
                    self._update_IS_events(recovery_event=existing_node)
        print(f"Number of already infected nodes: {infct_hd_nodes}, which is {(infct_hd_nodes/how_many)} percent of the projected vacine proportion")
        print(f"In total, {num_vaccinated} were vaccinated, out of a planned {how_many}, with max degree {degree}")

class TargetedRolloutSimulation(simulation.Simulation):
    def __init__(self, N, adjlist=None, max_unitless_sim_time=1000000, membership_groups=None, node_memberships=None):
        super().__init__(N=N, adj_list=adjlist, max_unitless_sim_time=max_unitless_sim_time,
                         membership_groups=membership_groups,
                         node_memberships=node_memberships)
        self.intervention_gen_list = None
        self.beta_redux_list = None
        self.proportion_reduced_list = None
        self.intervened_status_list = []
        self.time_of_intervention_list = []
        self.next_up_intervention_entry = 0
        self.use_uniform_rate = True

    def simtype(self):
        print('Simulation class of type Multi intervention Targeted Intervention')

    def configure_intervention(self, intervention_gen_list, beta_redux_list, proportion_reduced_list):
        self.intervention_gen_list = intervention_gen_list
        self.beta_redux_list = beta_redux_list
        self.proportion_reduced_list = proportion_reduced_list
        for i in range(len(intervention_gen_list)):
            self.intervened_status_list.append(False)

    def run_sim(self, with_memberships=False, uniform_rate=True, wait_for_recovery=False, visualize=False,
                viz_graph=None, viz_pos=None, p_zero=None, kill_by=None, record_active_gen_sizes=False):
        self.use_uniform_rate = uniform_rate
        if with_memberships: self.track_memberships = True
        self.use_uniform_rate = uniform_rate
        if self.track_memberships:
            self._init_membership_state_time_series()
        self._initialize_patient_zero()
        while self._current_sim_time < self.total_sim_time:
            if self.next_up_intervention_entry < len(self.intervention_gen_list):
                if not self.intervened_status_list[self.next_up_intervention_entry]:
                    if self._highest_gen >= self.intervention_gen_list[self.next_up_intervention_entry]:
                        self.intervene(self.next_up_intervention_entry)
                        self.time_of_intervention_list.append(self._current_sim_time)
                        self.intervened_status_list[self.next_up_intervention_entry] = True
                        self.next_up_intervention_entry += 1
            # Run one step
            self._single_step(uniform_rate=uniform_rate, visualize=visualize, viz_graph=viz_graph, viz_pos=viz_pos,
                              record_active_gen_sizes=record_active_gen_sizes)

            self._total_num_timesteps += 1
            if self._current_number_IS_events == 0:
                if not wait_for_recovery:
                    break
                elif len(self._recovery_events) == 0:
                    break
            if kill_by is not None and self._highest_gen >= kill_by:
                should_quit = self._check_active_gens(kill_by)
                if should_quit:
                    break

    def intervene(self, intervention_entry):
        if self._N == 0:
            self._N = len(self._adjlist)
        frac_of_network = self.proportion_reduced_list[intervention_entry] * self._N
        how_many = 0
        if frac_of_network > 1:
            how_many = int(np.round(frac_of_network, 0))

        degrees = list([len(neighbor) - 1 for neighbor in self._adjlist])
        degree_classes = {}
        for i in range(len(degrees)):
            degree = degrees[i]
            try:
                degree_classes[degree].append(i)
            except KeyError:
                degree_classes[degree] = [i]

        keys_decreasing = np.sort(list(degree_classes.keys()))[::-1]

        num_vaccinated = 0
        infct_hd_nodes = 0
        for degree in keys_decreasing:
            current_node_group = degree_classes[degree] # To match analytical prediction, once a degree class is bumped up to meet the quota, the whole class is vaccinated
            if num_vaccinated < how_many:
                for node in current_node_group:
                    if self.use_uniform_rate:
                        candidate_node = network.Node(node, -1, None, self._uniform_gamma)
                    else:
                        candidate_node = network.Node(node, -1, None, self._Gamma[node])
                    if self.track_memberships:
                        candidate_node.set_membership(self._node_memberships[candidate_node.get_label()])
                    existing_node = self._existing_node(candidate_node)
                    if existing_node.get_state() != nodestate.NodeState.VACCINATED:
                        if existing_node.get_state() != nodestate.NodeState.INFECTED: #Do not vaccinate currently infected nodes
                            existing_node.vaccinate()
                            num_vaccinated += 1 #Only counted if high degree nodes are successfully vaccinated
                        else:
                            infct_hd_nodes += 1 #Counts high-degree currently infected nodes
                    self._update_IS_events(recovery_event=existing_node)

class RingInterventionSim(simulation.Simulation):
    # TODO
    def simtype(self):
        print('I am a simulation class of type Ring Intervention')


class AbsoluteTimeNetworkSwitchSim(simulation.Simulation):
    def __init__(self, N, adjlist=None, max_unitless_sim_time=1000000, membership_groups=None, node_memberships=None):
        super().__init__(N=N, adj_list=adjlist, max_unitless_sim_time=max_unitless_sim_time,
                         membership_groups=membership_groups,
                         node_memberships=node_memberships)
        self.A_modified = None
        self.adjlist_modified = None
        self.intervention_time = None
        self.intervened = False
        self.time_of_intervention = max_unitless_sim_time
        self.new_Beta = None
        self.new_Gamma = None
        self.use_uniform_rate = True

    def configure_intervention(self, intervention_time, new_adjlist=None, new_beta=None, new_gamma=None):
        self.adjlist_modified = new_adjlist
        # self.A_modified = new_adjacency_matrix
        self.intervention_time = intervention_time
        self.new_Beta = new_beta
        self.new_Gamma = new_gamma

    def simtype(self):
        print('I am a simulation class of type Absolute Time Network Intervention')

    def run_sim(self, with_memberships=False, uniform_rate=True, wait_for_recovery=False, visualize=False,
                viz_graph=None, viz_pos=None, p_zero=None, kill_by=None, record_active_gen_sizes=False):
        self.use_uniform_rate = uniform_rate
        if with_memberships: self.track_memberships = True
        if self.track_memberships:
            self._init_membership_state_time_series()
        self._initialize_patient_zero(label=p_zero)
        while self._current_sim_time < self.total_sim_time:
            if not self.intervened:
                if self._current_sim_time > self.intervention_time:
                    self.intervene()
                    self.time_of_intervention = self._current_sim_time
                    self.intervened = True
            # Run one step
            self._single_step(uniform_rate=uniform_rate, visualize=visualize, viz_graph=viz_graph)

            self._total_num_timesteps += 1
            if self._current_number_IS_events == 0:
                if not wait_for_recovery:
                    break
                elif len(self._recovery_events) == 0:
                    break
            if kill_by is not None and self._highest_gen >= kill_by:
                should_quit = self._check_active_gens(kill_by)
                if should_quit:
                    break

    def intervene(self):
        if self.adjlist_modified is not None:
            self.set_adjlist(self.adjlist_modified)
        if self.new_Gamma is not None and self.new_Beta is not None:
            super().set_uniform_beta(self.new_Beta)
            super().set_uniform_gamma(self.new_Gamma)
        self._reset_node_states()
        for node_key in list(self._active_node_dict.keys()):
            node = self._active_node_dict[node_key]
            if node.get_state() == nodestate.NodeState.INFECTED:
                self._add_IS_events(node)
